#include "qe3.h"

visgroup_t	visgroups;
visgroup_t	*currentvisgroup;

void VisGroup_AddToList(visgroup_t *g)
{
	//add new group to the tail of the list
	g->next = &visgroups;
	g->prev = visgroups.prev;
	visgroups.prev->next = g;
	visgroups.prev = g;
}

void VisGroup_RemoveFromList(visgroup_t *g)
{
	g->next->prev = g->prev;
	g->prev->next = g->next;

	g->next = g->prev = NULL;
}

void VisGroups_Init(void)
{
	visgroup_t	*defaultgroup;

	visgroups.next = visgroups.prev = &visgroups;
	defaultgroup = malloc(sizeof(visgroup_t));
	defaultgroup->identifier = 0;
	defaultgroup->name = "default group";
	defaultgroup->color[0] = defaultgroup->color[1] = defaultgroup->color[2] = 0.0f;
	VisGroup_AddToList(defaultgroup);
	currentvisgroup = GetVisGroup(0);
}

int VisGroup_AddNew(char *name, vec3_t color, int identifier)
{
	visgroup_t	*group;

	for(group = visgroups.next; group != &visgroups; group = group->next)
	{
		if(group->identifier == identifier) // Vis groups must have a unique identifier
			return false;
	}

	group = malloc(sizeof(visgroup_t));
	group->identifier = identifier;
	group->name = strdup(name);
	VectorCopy(color, group->color);
	VisGroup_AddToList(group);

	return true;
}

int VisGroup_Renumber(void) // Called when a group is removed
{
	visgroup_t	*g;
	int			identifier = 0;

	for(g = visgroups.next; g != &visgroups; g = g->next)
		g->identifier = identifier++;
	return identifier;
}

void VisGroup_SetName(visgroup_t *g, char *newname)
{
	if(g->identifier == 0)
	{
		MessageBox(globals.d_hwndMain, "Cannot change name of default group", "Error", MB_OK | MB_ICONERROR );
		return;
	}

	free(g->name);
	g->name = strdup(newname);
}

void Visgroup_WriteHeader(FILE *f)
{
	visgroup_t	*g;
	int			numgroups;

	numgroups = VisGroup_Renumber(); // Make sure identifiers are contiguous

//	fprintf (f, "//QE5NumVisGroups = %d\n", numgroups);

	for(g = visgroups.next; g != &visgroups; g = g->next)
	{
//		if(g->identifier == 0) // Don't write default group
//			continue;

		fprintf(f, "//QE5VisGroupDef = %d %f %f %f %s\n", g->identifier, g->color[0], g->color[1], g->color[2], g->name);
	}
}

visgroup_t *GetVisGroup(int identifier)
{
	visgroup_t	*g;

	for(g = visgroups.next; g != &visgroups; g = g->next)
	{
		if(g->identifier == identifier)
			return g;
	}

	return visgroups.next; // return the default vis group
}

void ParseVisGroup(char **line)
{
	int		identifier;
	vec3_t	color;
	char	*current, *errmsg = "unknown setting";

//	Sys_Printf("Parsing Vis Group: %s\n", *line);
	if(!strncmp(*line, "//QE5VisGroupDef = ", 19))
	{
		errmsg = "vis group definition";
		get_token(line, "=");
		current = get_token(line, " ");
		if(!current)
			goto parsevisgroup_error;
		identifier = atoi(current);
		current = get_token(line, " ");
		if(!current)
			goto parsevisgroup_error;
		color[0] = atof(current);
		current = get_token(line, " ");
		if(!current)
			goto parsevisgroup_error;
		color[1] = atof(current);
		current = get_token(line, " ");
		if(!current)
			goto parsevisgroup_error;
		color[2] = atof(current);

//		if(**line != '')
		if(!**line || !*line)
			goto parsevisgroup_error;

//		for(i = 0; *line[i] != '\n'; i++)
//		{
//			if(*line[i] == '')
//			{
//				*line[i] = 0;

				// default vis group's name doesn't change, just set the color
				if(!identifier)
				{
					VectorCopy(color, visgroups.next->color);
				}
				else
					VisGroup_AddNew(*line, color, identifier);

//				*line = &*line[i+1];
				return;
//			}
//		}
	}
	if(!strncmp(*line, "//QE5VisGroup = ", 16))
	{
		errmsg = "vis group";
		get_token(line, "=\n");
		current = get_token(line, " ");

		if(!current)
			goto parsevisgroup_error;

		identifier = atoi(current);
		currentvisgroup = GetVisGroup(identifier);
		return;
	}

parsevisgroup_error:
	Sys_Printf("Error parsing %s\n", errmsg);
}

void ReplaceVisGroup(visgroup_t *old, visgroup_t *newgroup)
{
	brush_t	*b;

	for(b = active_brushes.next; b != &active_brushes; b = b->next)
	{
		if(b->group == old)
			b->group = newgroup;
	}
	for(b = selected_brushes.next; b != &selected_brushes; b = b->next)
	{
		if(b->group == old)
			b->group = newgroup;
	}
	for(b = filtered_brushes.next; b != &filtered_brushes; b = b->next)
	{
		if(b->group == old)
			b->group = newgroup;
	}
}

void VisGroup_Free(visgroup_t *g)
{
	VisGroup_RemoveFromList(g);
	free(g->name);
	free(g);
}

void VisGroup_FillListBox(HWND h)
{
	visgroup_t	*g;

	VisGroup_Renumber();
	
	SendMessage(h, LB_RESETCONTENT, 0 , 0);
	
	for(g = visgroups.next; g != &visgroups; g = g->next)
	{
		SendMessage(h, LB_ADDSTRING, 0, (LPARAM)g->name);
	}
	
	SendMessage(h, LB_SETSEL, TRUE, 0);
}

void VisGroup_SetSelected(void)
{
	brush_t *b;

	for(b = selected_brushes.next; b != &selected_brushes; b = b->next)
		b->group = currentvisgroup;
}

void VisGroup_ClearAll(void)
{
	visgroup_t	*group, *gnext;

	for(group = visgroups.next; group != &visgroups; group = gnext)
	{
		gnext = group->next;
		if(group->identifier == 0)
			continue;

		VisGroup_Free(group);
	}
}
